MODULE traldf_iso
   !!==============================================================================
   !!                    ***  MODULE  traldf_iso  ***
   !! Ocean active tracers:  horizontal component of the lateral tracer mixing trend
   !!==============================================================================
#if defined key_ldfslp   ||   defined key_esopa
   !!----------------------------------------------------------------------
   !!   'key_ldfslp'                  rotation of the lateral mixing tensor
   !!----------------------------------------------------------------------
   !!   tra_ldf_iso : update the tracer trend with the horizontal component
   !!                 of iso neutral laplacian operator or horizontal 
   !!                 laplacian operator in s-coordinate
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers variables
   USE dom_oce         ! ocean space and time domain variables
   USE ldftra_oce     ! ocean active tracers: lateral physics
   USE trdtra_oce     ! ocean active tracers: trend variables
   USE in_out_manager  ! I/O manager
   USE ldfslp          ! iso-neutral slopes
   USE lbclnk

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC tra_ldf_iso  ! routine called by step.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "ldftra_substitute.h90"
#  include "ldfeiv_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE tra_ldf_iso( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_ldf_iso  ***
      !! 
      !! ** Purpose :   Compute the before horizontal tracer (t & s) diffusive 
      !!      trend and add it to the general trend of tracer equation.
      !!
      !! ** Method  :   The horizontal component of the lateral diffusive trends 
      !!      is provided by a 2nd order operator rotated along neural or geopo-
      !!      tential surfaces to which an eddy induced advection can be added
      !!      It is computed using before fields (forward in time) and isopyc-
      !!      nal or geopotential slopes computed in routine ldfslp.
      !!
      !!      horizontal fluxes associated with the rotated lateral mixing:
      !!         zftu = (aht+ahtb0) e2u*e3u/e1u di[ tb ]
      !!               - aht       e2u*uslp    dk[ mi(mk(tb)) ]
      !!         zftv = (aht+ahtb0) e1v*e3v/e2v dj[ tb ]
      !!               - aht       e2u*vslp    dk[ mj(mk(tb)) ]
      !!      add horizontal Eddy Induced advective fluxes (lk_traldf_eiv=T):
      !!         zftu = zftu - dk-1[ aht e2u mi(wslpi) ] mi( tb ) 
      !!         zftv = zftv - dk-1[ aht e1v mj(wslpj) ] mj( tb ) 
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1t*e2t*e3t) {  di-1[ zftu ] +  dj-1[ zftv ]  }
      !!      Add this trend to the general trend (ta,sa):
      !!         ta = ta + difft
      !!
      !! ** Action  : - Update (ta,sa) arrays with the before isopycnal or
      !!                geopotential s-coord harmonic mixing trend.
      !!              - Save the trends in (ttrd,strd) ('key_diatrends')
      !!
      !! History :
      !!        !  94-08  (G. Madec, M. Imbard)
      !!        !  97-05  (G. Madec)  split into traldf and trazdf
      !!   8.5  !  02-08  (G. Madec)  Free form, F90
      !!----------------------------------------------------------------------
      !! * Modules used
      USE oce           , zftu => ua,  &  ! use ua as workspace
         &                zfsu => va      ! use va as workspace

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt       ! ocean time-step index

      !! * Local declarations
      INTEGER ::   ji, jj, jk             ! dummy loop indices
      REAL(wp) ::   &
         zabe1, zabe2, zcof1, zcof2,   &  ! temporary scalars
         zmsku, zmskv, zbtr, zta, zsa, &
#if defined key_traldf_eiv
         zcg1, zcg2, zuwk, zvwk,       &
         zuwk1, zvwk1,                 &
#endif
         ztagu, ztagv, zsagu, zsagv
      REAL(wp), DIMENSION(jpi,jpj) ::   &
         zdkt, zdk1t,      &  ! workspace
         zdks, zdk1s
#if defined key_traldf_eiv
      REAL(wp), DIMENSION(jpi,jpj) ::   &
         zftug, zftvg,                  &
         zfsug, zfsvg
#endif
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &
         zftv, zfsv                       ! workspace
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_ldf_iso : iso neutral lateral diffusion or'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~   horizontal laplacian diffusion in s-coordinate'
#if defined key_diaeiv
         u_eiv(:,:,:) = 0.e0
         v_eiv(:,:,:) = 0.e0
#endif
      ENDIF

      ztagu = 0.e0
      ztagv = 0.e0
      zsagu = 0.e0
      zsagv = 0.e0

      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
         ! 1. Vertical tracer gradient at level jk and jk+1
         ! ------------------------------------------------
         ! surface boundary condition: zdkt(jk=1)=zdkt(jk=2)

         zdk1t(:,:) = ( tb(:,:,jk) - tb(:,:,jk+1) ) * tmask(:,:,jk+1)
         zdk1s(:,:) = ( sb(:,:,jk) - sb(:,:,jk+1) ) * tmask(:,:,jk+1)

         IF( jk == 1 ) THEN
            zdkt(:,:) = zdk1t(:,:)
            zdks(:,:) = zdk1s(:,:)
         ELSE
            zdkt(:,:) = ( tb(:,:,jk-1) - tb(:,:,jk) ) * tmask(:,:,jk)
            zdks(:,:) = ( sb(:,:,jk-1) - sb(:,:,jk) ) * tmask(:,:,jk)
         ENDIF


         ! 2. Horizontal fluxes
         ! --------------------

         DO jj = 1 , jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zabe1 = ( fsahtu(ji,jj,jk) + ahtb0 ) * e2u(ji,jj) * fse3u(ji,jj,jk) / e1u(ji,jj)
               zabe2 = ( fsahtv(ji,jj,jk) + ahtb0 ) * e1v(ji,jj) * fse3v(ji,jj,jk) / e2v(ji,jj)

               zmsku = 1. / MAX(   tmask(ji+1,jj,jk  ) + tmask(ji,jj,jk+1)   &
                                 + tmask(ji+1,jj,jk+1) + tmask(ji,jj,jk  ), 1. )

               zmskv = 1. / MAX(   tmask(ji,jj+1,jk  ) + tmask(ji,jj,jk+1)   &
                                 + tmask(ji,jj+1,jk+1) + tmask(ji,jj,jk  ), 1. )

               zcof1 = -fsahtu(ji,jj,jk) * e2u(ji,jj) * uslp(ji,jj,jk) * zmsku
               zcof2 = -fsahtv(ji,jj,jk) * e1v(ji,jj) * vslp(ji,jj,jk) * zmskv

               zftu(ji,jj,jk) = umask(ji,jj,jk) * (   zabe1 * (   tb(ji+1,jj,jk) - tb(ji,jj,jk)  )   &
                  &                              + zcof1 * (   zdkt (ji+1,jj) + zdk1t(ji,jj)      &
                  &                                          + zdk1t(ji+1,jj) + zdkt (ji,jj)  )  )

               zftv(ji,jj,jk) = vmask(ji,jj,jk) * (   zabe2 * (   tb(ji,jj+1,jk) - tb(ji,jj,jk)  )   &
                  &                              + zcof2 * (   zdkt (ji,jj+1) + zdk1t(ji,jj)      &
                  &                                          + zdk1t(ji,jj+1) + zdkt (ji,jj)  )  )

               zfsu(ji,jj,jk) = umask(ji,jj,jk) * (   zabe1 * (   sb(ji+1,jj,jk) - sb(ji,jj,jk)  )   &
                  &                              + zcof1 * (   zdks (ji+1,jj) + zdk1s(ji,jj)      &
                  &                              + zdk1s(ji+1,jj) + zdks (ji,jj)  )  )

               zfsv(ji,jj,jk) = vmask(ji,jj,jk) * (   zabe2 * (   sb(ji,jj+1,jk) - sb(ji,jj,jk)  )   &
                  &                              + zcof2 * (   zdks (ji,jj+1) + zdk1s(ji,jj)      &
                  &                              + zdk1s(ji,jj+1) + zdks (ji,jj)  )  )
            END DO
         END DO

#   if defined key_traldf_eiv
         !                              ! ---------------------------------------!
         !                              ! Eddy induced vertical advective fluxes !
         !                              ! ---------------------------------------!
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  zuwk = ( wslpi(ji,jj,jk  ) + wslpi(ji+1,jj,jk  ) ) * fsaeiu(ji,jj,jk  ) * umask(ji,jj,jk  )
                  zuwk1= ( wslpi(ji,jj,jk+1) + wslpi(ji+1,jj,jk+1) ) * fsaeiu(ji,jj,jk+1) * umask(ji,jj,jk+1)
                  zvwk = ( wslpj(ji,jj,jk  ) + wslpj(ji,jj+1,jk  ) ) * fsaeiv(ji,jj,jk  ) * vmask(ji,jj,jk  )
                  zvwk1= ( wslpj(ji,jj,jk+1) + wslpj(ji,jj+1,jk+1) ) * fsaeiv(ji,jj,jk+1) * vmask(ji,jj,jk+1)
   
                  zcg1= -0.25 * e2u(ji,jj) * umask(ji,jj,jk) * ( zuwk-zuwk1 )
                  zcg2= -0.25 * e1v(ji,jj) * vmask(ji,jj,jk) * ( zvwk-zvwk1 )
   
                  zftug(ji,jj) = zcg1 * ( tb(ji+1,jj,jk) + tb(ji,jj,jk) )
                  zftvg(ji,jj) = zcg2 * ( tb(ji,jj+1,jk) + tb(ji,jj,jk) )
                  zfsug(ji,jj) = zcg1 * ( sb(ji+1,jj,jk) + sb(ji,jj,jk) )
                  zfsvg(ji,jj) = zcg2 * ( sb(ji,jj+1,jk) + sb(ji,jj,jk) )
   
                  zftu(ji,jj,jk) = zftu(ji,jj,jk) + zftug(ji,jj)
                  zftv(ji,jj,jk) = zftv(ji,jj,jk) + zftvg(ji,jj)
                  zfsu(ji,jj,jk) = zfsu(ji,jj,jk) + zfsug(ji,jj)
                  zfsv(ji,jj,jk) = zfsv(ji,jj,jk) + zfsvg(ji,jj)
#   if defined key_diaeiv
                  u_eiv(ji,jj,jk) = -2. * zcg1 / ( e2u(ji,jj) * fse3u(ji,jj,jk) )
                  v_eiv(ji,jj,jk) = -2. * zcg2 / ( e1v(ji,jj) * fse3v(ji,jj,jk) )
#   endif
               END DO
            END DO
#   endif

         ! II.4 Second derivative (divergence) and add to the general trend
         ! ----------------------------------------------------------------

         DO jj = 2 , jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zbtr= 1. / ( e1t(ji,jj)*e2t(ji,jj)*fse3t(ji,jj,jk) )
               zta = zbtr * (  zftu(ji,jj,jk) - zftu(ji-1,jj  ,jk)   &
                  &          + zftv(ji,jj,jk) - zftv(ji  ,jj-1,jk)  )
               zsa = zbtr * (  zfsu(ji,jj,jk) - zfsu(ji-1,jj  ,jk)   &
                  &          + zfsv(ji,jj,jk) - zfsv(ji  ,jj-1,jk)  )
               ta (ji,jj,jk) = ta (ji,jj,jk) + zta
               sa (ji,jj,jk) = sa (ji,jj,jk) + zsa
#if defined key_trdtra || defined key_trdmld
#   if defined key_traldf_eiv
               ztagu = ( zftug(ji,jj) - zftug(ji-1,jj  ) ) * zbtr
               ztagv = ( zftvg(ji,jj) - zftvg(ji  ,jj-1) ) * zbtr
               zsagu = ( zfsug(ji,jj) - zfsug(ji-1,jj  ) ) * zbtr
               zsagv = ( zfsvg(ji,jj) - zfsvg(ji  ,jj-1) ) * zbtr
               ttrdh(ji,jj,jk,3) = ztagu
               ttrdh(ji,jj,jk,4) = ztagv
               strdh(ji,jj,jk,3) = zsagu
               strdh(ji,jj,jk,4) = zsagv
#   endif
               ttrd (ji,jj,jk,3) = zta - ztagu - ztagv
               strd (ji,jj,jk,3) = zsa - zsagu - zsagv
#endif
            END DO
         END DO
         !                                          ! ===============
      END DO                                        !   End of slab  
      !                                             ! ===============

      IF( l_ctl .AND. lwp ) THEN         ! print mean trends (used for debugging)
         zta = SUM( ta(2:jpim1,2:jpjm1,1:jpkm1) * tmask(2:jpim1,2:jpjm1,1:jpkm1) )
         zsa = SUM( sa(2:jpim1,2:jpjm1,1:jpkm1) * tmask(2:jpim1,2:jpjm1,1:jpkm1) )
         WRITE(numout,*) ' ldf  - Ta: ', zta-t_ctl, ' Sa: ', zsa-s_ctl
         t_ctl = zta   ;   s_ctl = zsa
      ENDIF

#if defined key_diaptr
      !!bug  no separation of diff iso and eiv
      IF( MOD( kt, nf_ptr ) == 0 ) THEN
      ! "zonal" mean lateral diffusive heat and salt transports
         pht_ldf(:,:) = prt_vj( zftv(:,:,:) )
         pst_ldf(:,:) = prt_vj( zfsv(:,:,:) )
         ! "zonal" mean lateral eddy induced velocity heat and salt transports
         pht_eiv(:,:) = prt_vj( zftv(:,:,:) )
         pst_eiv(:,:) = prt_vj( zfsv(:,:,:) )
      ENDIF
#endif

   END SUBROUTINE tra_ldf_iso

#else
   !!----------------------------------------------------------------------
   !!   Dummy module :             No rotation of the lateral mixing tensor
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE tra_ldf_iso( kt )               ! Empty routine
      WRITE(*,*) 'tra_ldf_iso: You should not have seen this print! error?', kt
   END SUBROUTINE tra_ldf_iso
#endif

   !!==============================================================================
END MODULE traldf_iso
